@@ -12,9 +12,8 @@ import pytest
 import stat
 import shutil
 import filecmp
+import time
 import errno
-import platform
-from distutils.version import LooseVersion
 from tempfile import NamedTemporaryFile
 from util import (wait_for_mount, umount, cleanup, base_cmdline,
                   safe_sleep, basename, fuse_test_marker)
@@ -86,11 +85,13 @@ def test_passthrough(tmpdir, name, debug, capfd):
 
         tst_statvfs(work_dir)
         tst_readdir(src_dir, work_dir)
+        tst_open_read(src_dir, work_dir)
+        tst_open_write(src_dir, work_dir)
+        tst_create(work_dir)
+        tst_passthrough(src_dir, work_dir)
         if not is_ll:
             tst_mkdir(work_dir)
             tst_rmdir(src_dir, work_dir)
-            tst_create(work_dir)
-            tst_open_write(src_dir, work_dir)
             tst_unlink(src_dir, work_dir)
             tst_symlink(work_dir)
             if os.getuid() == 0:
@@ -103,7 +104,6 @@ def test_passthrough(tmpdir, name, debug, capfd):
             tst_truncate_path(work_dir)
             tst_truncate_fd(work_dir)
             tst_open_unlink(work_dir)
-            tst_passthrough(src_dir, work_dir)
 
             subprocess.check_call([ os.path.join(basename, 'test', 'test_syscalls'),
                                     work_dir, ':' + src_dir ])
@@ -176,36 +176,6 @@ def test_null(tmpdir):
         umount(mount_process, mnt_file)
 
 
-@pytest.mark.parametrize("name",
-                         ('notify_inval_inode',
-                          'notify_store_retrieve'))
-@pytest.mark.parametrize("notify", (True, False))
-def test_notify1(tmpdir, name, notify):
-    mnt_dir = str(tmpdir)
-    cmdline = base_cmdline + \
-              [ pjoin(basename, 'example', name),
-                '-f', '--update-interval=1', mnt_dir ]
-    if not notify:
-        cmdline.append('--no-notify')
-    mount_process = subprocess.Popen(cmdline)
-    try:
-        wait_for_mount(mount_process, mnt_dir)
-        filename = pjoin(mnt_dir, 'current_time')
-        with open(filename, 'r') as fh:
-            read1 = fh.read()
-        safe_sleep(2)
-        with open(filename, 'r') as fh:
-            read2 = fh.read()
-        if notify:
-            assert read1 != read2
-        else:
-            assert read1 == read2
-    except:
-        cleanup(mnt_dir)
-        raise
-    else:
-        umount(mount_process, mnt_dir)
-
 @pytest.mark.parametrize("notify", (True, False))
 def test_notify_inval_entry(tmpdir, notify):
     mnt_dir = str(tmpdir)
@@ -239,21 +209,6 @@ def test_notify_inval_entry(tmpdir, notify):
     else:
         umount(mount_process, mnt_dir)
 
-@pytest.mark.parametrize("writeback", (False, True))
-def test_write_cache(tmpdir, writeback):
-    if writeback and LooseVersion(platform.release()) < '3.14':
-        pytest.skip('Requires kernel 3.14 or newer')
-    # This test hangs under Valgrind when running close(fd)
-    # test_write_cache.c:test_fs(). Most likely this is because of an internal
-    # deadlock in valgrind, it probably assumes that until close() returns,
-    # control does not come to the program.
-    mnt_dir = str(tmpdir)
-    cmdline = [ pjoin(basename, 'test', 'test_write_cache'),
-                mnt_dir ]
-    if writeback:
-        cmdline.append('-owriteback_cache')
-    subprocess.check_call(cmdline)
-
 @pytest.mark.skipif(os.getuid() != 0,
                     reason='needs to run as root')
 def test_cuse(capfd):
@@ -371,6 +326,14 @@ def tst_chown(mnt_dir):
     assert fstat.st_uid == uid_new
     assert fstat.st_gid == gid_new
 
+def tst_open_read(src_dir, mnt_dir):
+    name = name_generator()
+    with open(pjoin(src_dir, name), 'wb') as fh_out, \
+         open(TEST_FILE, 'rb') as fh_in:
+        shutil.copyfileobj(fh_in, fh_out)
+
+    assert filecmp.cmp(pjoin(mnt_dir, name), TEST_FILE, False)
+
 def tst_open_write(src_dir, mnt_dir):
     name = name_generator()
     fd = os.open(pjoin(src_dir, name),
@@ -407,33 +370,56 @@ def tst_link(mnt_dir):
     name2 = pjoin(mnt_dir, name_generator())
     shutil.copyfile(TEST_FILE, name1)
     assert filecmp.cmp(name1, TEST_FILE, False)
+
+    fstat1 = os.lstat(name1)
+    assert fstat1.st_nlink == 1
+
     os.link(name1, name2)
 
     fstat1 = os.lstat(name1)
     fstat2 = os.lstat(name2)
-
     assert fstat1 == fstat2
     assert fstat1.st_nlink == 2
-
     assert os.path.basename(name2) in os.listdir(mnt_dir)
     assert filecmp.cmp(name1, name2, False)
+
+    # Since RELEASE requests are asynchronous, it is possible that
+    # libfuse still considers the file to be open at this point
+    # and (since -o hard_remove is not used) renames it instead of
+    # deleting it. In that case, the following lstat() call will
+    # still report an st_nlink value of 2 (cf. issue #157).
     os.unlink(name2)
+
+    assert os.path.basename(name2) not in os.listdir(mnt_dir)
+    with pytest.raises(FileNotFoundError):
+        os.lstat(name2)
+
+    # See above, we may have to wait until RELEASE has been
+    # received before the st_nlink value is correct.
+    maxwait = time.time() + 2
     fstat1 = os.lstat(name1)
+    while fstat1.st_nlink == 2 and time.time() < maxwait:
+        fstat1 = os.lstat(name1)
+        time.sleep(0.1)
     assert fstat1.st_nlink == 1
+
     os.unlink(name1)
 
 def tst_readdir(src_dir, mnt_dir):
-    dir_ = pjoin(src_dir, name_generator())
-    file_ = dir_ + "/" + name_generator()
-    subdir = dir_ + "/" + name_generator()
+    newdir = name_generator()
+
+    src_newdir = pjoin(src_dir, newdir)
+    mnt_newdir = pjoin(mnt_dir, newdir)
+    file_ = src_newdir + "/" + name_generator()
+    subdir = src_newdir + "/" + name_generator()
     subfile = subdir + "/" + name_generator()
 
-    os.mkdir(dir_)
+    os.mkdir(src_newdir)
     shutil.copyfile(TEST_FILE, file_)
     os.mkdir(subdir)
     shutil.copyfile(TEST_FILE, subfile)
 
-    listdir_is = os.listdir(dir_)
+    listdir_is = os.listdir(mnt_newdir)
     listdir_is.sort()
     listdir_should = [ os.path.basename(file_), os.path.basename(subdir) ]
     listdir_should.sort()
@@ -442,7 +428,7 @@ def tst_readdir(src_dir, mnt_dir):
     os.unlink(file_)
     os.unlink(subfile)
     os.rmdir(subdir)
-    os.rmdir(dir_)
+    os.rmdir(src_newdir)
 
 def tst_truncate_path(mnt_dir):
     assert len(TEST_DATA) > 1024
